import java.io.*;
import java.util.*;
import java.util.Map.Entry;


class StationInfo{

    private String stationName;
    private String lineName;

    public StationInfo(String stationName, String lineName){

        this.stationName=stationName;
        this.lineName=lineName;
    }

    public String getStationName(){
        return stationName;
    }

    public String getLineName(){
        return lineName;
    }

}

public class SubwayMap {

    private Map<String, StationInfo> idMap; //key : 역의 id, value : 역의 이름, 호선의 이름
    private Map<String, LinkedList<Edge>> stationMap; //key : 역의 이름, value : edgelist
    private Map<String, Integer> transitTime;


    private static final int INF = 100000000; //아직 역 사이로 이동이 불가함을 나타내는 정수
    private static final int TRANSITTIME_DEFAULT = 5; //환승 시간

    //지하철 노선도의 생성자
    public SubwayMap(){
        this.idMap = new HashMap<>();
        this.stationMap = new HashMap<>();
        this.transitTime=new HashMap<>();
    }

    //역의 아이디를 역의 이름으로 바꿔주는 함수
    private String id2name(String stationID){
        return this.idMap.get(stationID).getStationName();
    }

    //역의 아이디를 역의 이름과 역의 호선을 포함한 정보로 바꿔주는 함수(역의 아이디에 저장된 호선을 반환)
    private StationInfo id2info(String stationID){
        return this.idMap.get(stationID);
    }

    private String toTransfer(String stationName){
        return "[" + stationName + "]";
    }

    //역 정보를 받아서 stationMap에 저장하는 함수
    public void insertStation(String stationID, String stationName, String limeName){

        if(stationID.contains("-")) limeName+="BL";
        StationInfo stationInfo = new StationInfo(stationName, limeName);

        idMap.put(stationID, stationInfo);
        stationMap.putIfAbsent(stationName, new LinkedList<Edge>());

    }

    //간선 정보를 받아서 edgelist에 저장하는 함수
    public void insertEdge(String departureID, String destinationID, int time){

        String departureName = this.id2name(departureID); //출발역 이름
        StationInfo destinationInfo = this.id2info(destinationID); //도착역의 정보
        String destinationName = destinationInfo.getStationName(); //도착역 이름
        String destinationLine = destinationInfo.getLineName(); //도착역 호선 이름

        Edge edge = new Edge(time, destinationName, destinationLine); //간선 만들기
        stationMap.get(departureName).add(edge); //
    }

    public void insertTransittime(String stationName, int time) {
        transitTime.put(stationName, time);
    }

    //역 이름을 넣어서 역에 연결된 모든 edge를 list로 가지는 연결리스트 반환하는 함수
    private LinkedList<Edge> getEdgelist(String stationName){
        return stationMap.get(stationName);
    }

    public String[] displayRoute(String departureName, String destinationName) throws Exception{

        RouteItem route = getShortestRoute(departureName, destinationName);
        StringBuilder stringBuilder = route.routeList;

        String Route = new String(stringBuilder.deleteCharAt(stringBuilder.length()-1));
        String shortestTime = Integer.toString(route.totalTime);

        String[] returnString = new String[2];
        returnString[0] = Route;
        returnString[1] = shortestTime;

        return returnString;
    }

    private RouteItem getShortestRoute(String departureName, String destinationName) throws Exception{

        if(!stationMap.containsKey(departureName) || !stationMap.containsKey(destinationName)) {
            System.out.println("출발역과 도착역의 정보가 없습니다.");
            throw new Exception();
        }

        PriorityQueue<Qitem> stationList = new PriorityQueue<>(); //가능한 경로들을 저장하는 우선순위 큐(힙)
        Set<String> visitedStation = new HashSet<>(); //방문한 역의 이름을 저장하는 hashMap(지금까지 최단거리를 구한 모든 set)

        LinkedList<Edge> departure_Edgelist = this.getEdgelist(departureName); //처음 역의 연결된 모든 역을 저장한 list 얻기

        for(Edge edge : departure_Edgelist){

            visitedStation.add(departureName + edge.getLineName()); //출발역을 집합에 포함하기
            String stationName = edge.getDestinationName(); //edge에 저장되어 있는 역 이름
            StringBuilder current_routeList = new StringBuilder(departureName + " " + stationName + " ");
            stationList.add(edge.edge2Qitem(new RouteItem(current_routeList, edge.getTimeTaken()))); //큐에 edge에 저장되어 있는 정보를 이용하여 역정보 집어넣기(처음 출발하는 것이므로 누적된 time이 0이다.)

        }
        int repeat = 0;//시험용

        //우선순위 큐의 size가 0이 될 때까지 수행하기
        while(!stationList.isEmpty()){

            Qitem qitem = stationList.poll(); //우선순위 큐에서 최소인 것 반환(qitem)

            if(qitem.stationName.equals(destinationName)){
                return qitem.routeItem;
            }

            String polledStation = qitem.stationName;
            LinkedList<Edge> edgeList = getEdgelist(polledStation); // qitem -> edgelist

            visitedStation.add(polledStation + qitem.pastLineName); //최소인 것 방문처리하면서 집합에 포함하

            for(Edge edge : edgeList){
                String currentDestination=edge.getDestinationName();

                //edge가 향하는 목적지가 visited set에 포함되지 않았을 때 edge 정보를 우선순위 큐에 저장해야 된다.(단, 현재 queue에 있는 경우 그거보다 최단거리로 갱신해야됨.)
                if(!visitedStation.contains(currentDestination + edge.getLineName())){

                    int isDifferentline = edge.getLineName().compareTo(qitem.pastLineName);
                    int toedgeTime = qitem.routeItem.totalTime + edge.getTimeTaken();
                    StringBuilder goThrough = qitem.routeItem.RouteItem2string();

                    if(isDifferentline != 0){

                        if(transitTime.containsKey(qitem.stationName)){
                            toedgeTime+=transitTime.get(qitem.stationName);
                        }
                        else {
                            toedgeTime += TRANSITTIME_DEFAULT;
                        }
                        goThrough.insert(goThrough.length()-polledStation.length()-1, '[');
                        goThrough.insert(goThrough.length()-1, ']');
                    }

                    goThrough.append(currentDestination + " ");
                    stationList.add(edge.edge2Qitem(new RouteItem(goThrough, toedgeTime)));
                }
            }
        }
        return null;
    }
}
